/*****************************************************************************************
 *
 * main.c
 * Copyright (c) 2010 NuukiaWorld
 *
 *  This file is part of "Morse tone".
 *
 *  "Network database" is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  "Network database" is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with "Morse tone".  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Version    Date          Author             Log
 *  --------------------------------------------------------------------------------------
 *  1.0        04.10.2010    NuukiaWorld        Initial Release
 *
 ****************************************************************************************/

#include "main.h"

/**************************************************************************************/

static struct {
   unsigned char  *ptr;                   // Points to <data>
   unsigned int   params;                 // play_tone() processes only these two first fields
   unsigned char  data[TONE_LEN];
} *tone;                                  // 4 static bytes reserved for pointer to generated tone

static const unsigned char tone_begin[] =  { 0x00, 0x02, 0xFD, 0x09, 0x0A, 0x80 };
static const unsigned char tone_end[] =    { 0x09, 0x0A, 0x7F, 0x0B };

// Add one command to the tone:
#define  TONE_CMD(CMD,DUR)                ( tone_ptr[0]=(CMD), tone_ptr[1]=(DUR), tone_ptr+=2 )
// Add two commands to the tone:
#define  TONE_CMD2(CMD1,DUR1,CMD2,DUR2)   ( tone_ptr[0]=(CMD1), tone_ptr[1]=(DUR1), \
                                            tone_ptr[2]=(CMD2), tone_ptr[3]=(DUR2), tone_ptr+=4 )

#define  NOTE  tone_sms_ascending_note    // Tone command used for morse (same as in "Ascending")
#define  PAUSE 0x40                       // Tone command used for silence
#define  VIBRA 0x0A                       // Tone command used for vibra

/**************************************************************************************/

/* Add one morse letter from the table to the tone
   (The tone must already end with a pause command if space is to be added)
   Returns updated <tone_ptr>
   
   Compressed table format (note: bits reversed, starting from LSB):
      000xxx   repeat xxx 2 times ( xxx != 000 )
      100x     add x (dit: x=0, dah: x=1)
      010x     repeat x 2 times
      110x     repeat x 3 times
      001x     repeat x 4 times
      101x     repeat x 5 times
      011xx    repeat xx 2 times
      111xx    repeat xx 3 times
*/
static unsigned char *add_letter( unsigned char *tone_ptr, char_t ch,
   unsigned int ticks_per_unit ) {

   unsigned int code, len, count;

   code = morse_table[ch];
   if( code != 0 ) {
      do {
         // Parse next compressed command
         count = code & 0x07;
         if( count == 0 ) {
            len = 6;          // repeat xxx
            count = 2;
         } else if( count <= 5 ) {
            len = 4;          // repeat x
         } else {
            len = 5;          // repeat xx
            count -= 4;
         }
         
         // Add corresponding tone commands
         if( count > 1 ) TONE_CMD( 0x05, count );                         // Start repeat
         // First symbol + symbol space
         TONE_CMD2( NOTE, ( code & 0x08 ? 3 : 1 ) * ticks_per_unit, PAUSE, ticks_per_unit );
         if( len > 4 ) {
            // Second symbol + symbol space
            TONE_CMD2( NOTE, ( code & 0x10 ? 3 : 1 ) * ticks_per_unit, PAUSE, ticks_per_unit );
            if( len > 5 ) {
               // Third symbol + symbol space
               TONE_CMD2( NOTE, ( code & 0x20 ? 3 : 1 ) * ticks_per_unit, PAUSE, ticks_per_unit );
            }
         }
         if( count > 1 ) *tone_ptr++ = 0x06;                              // End repeat
         
         code >>= len;
      } while( code != 0 );

      // Add letter space (2 extra units = 3 units)
      if( count > 1 ) { // previous pause cannot be extended because of repeat
         TONE_CMD( PAUSE, 2 * ticks_per_unit );
      } else {
         tone_ptr[-1] = 3 * ticks_per_unit;
      }
   } else {
      // Add word space (4 more extra units = 7 units), skip if already added
      if( tone_ptr[-1] <= 3 * ticks_per_unit ) tone_ptr[-1] += 4 * ticks_per_unit;
   }
   return tone_ptr;
}

/**************************************************************************************/

/* Appends 7-bit GSM alphabet text to tone as morse code
   Set skip<0 to create a preview tone (no CT, no AR)
   Returns updated <tone_ptr>
*/
static unsigned char *parse_text( unsigned char *tone_ptr, const unsigned char *text,
   unsigned int len, int skip ) {

   unsigned int shift, ticks_per_unit;
   char_t ch;
   
   // Get time for one morse "unit"
   ticks_per_unit = own_get_config(confid_ticks_per_unit);

   // Message start
   if( skip >= 0 ) {
      TONE_CMD( VIBRA, 0x01 );                                       // Vibra on
      tone_ptr = add_letter( tone_ptr, '\x80', ticks_per_unit );     // Code CT "start"
      tone_ptr -= 2;                                                 // Remove letter space
      TONE_CMD( VIBRA, 0xFE );                                       // Vibra off
   }
   TONE_CMD( PAUSE, 7 * ticks_per_unit );                            // Pause before text

   // Message contents
   shift = 25;
   do {
      if( len == 0 ) {
         // Message end
         if( skip >= 0 ) {
            tone_ptr = add_letter( tone_ptr, ' ', ticks_per_unit );  // Pause after text
            tone_ptr = add_letter( tone_ptr, '+', ticks_per_unit );  // Code AR "stop"
         }
         break;
      }
      
      if( skip <= 0 ) {
         // Extract next 7-bit character from the text ( expect: sizeof(unsigned int)==32bits )
         ch = (unsigned int)( text[1] << 8 | text[0] ) << shift >> 25;
         // Add character to the tone
         tone_ptr = add_letter( tone_ptr, ch, ticks_per_unit );
      } else {
         // Skipping <skip> initial characters (User Data Header)
         skip--;
      }
      
      shift -= 7;
      if( shift <= 17 ) {
         shift += 8;
         text++;
      }
      len--;
   } while( tone_ptr <= tone->data + ( TONE_LEN - TONE_LEN_LONGEST_CODE - sizeof(tone_end) ) );
      
   return tone_ptr;
}

/* Create SMS tone based on the message contents
   Set skip<0 to create a preview tone (no CT, no AR, no repeat)
   Returns non-zero if successful
*/
bool_t build_morse_tone( const unsigned char *text, unsigned int len, int skip ) {

   unsigned char *tone_ptr;

   if( tone == NULL ) {
      tone = malloc_try(sizeof(*tone));
      if( tone == NULL ) return false;                   // quit, out of memory
      tone->ptr = NULL;                                  // "Not playing"
      tone->params = tone_sms_ascending_params;
   } else if( tone->ptr != NULL ) {
      return false;                                      // quit, still playing previous
   }
   tone_ptr = tone->data;
   
   // Add tone beginning
   memcpy( tone_ptr, tone_begin, sizeof(tone_begin) );   // Technical tone start commands
   tone_ptr += sizeof(tone_begin);
   
   do {
      tone_ptr = parse_text( tone_ptr, text, len, skip );
      if( skip < 0 ) break;                              // No repeat in preview
      tone_ptr[-1] = TICKS_PER_SECOND * 2;               // Pause before repeat (2 sec)
      // Repeat the message if still less than 50% of the tone buffer used
   } while( tone_ptr <= tone->data + TONE_LEN / 2 );

   // Add tone end
   return (bool_t)memcpy( tone_ptr, tone_end, sizeof(tone_end) ); // Technical tone end commands
   // Non-zero memcpy() return value reused as "true" to save 2...4 bytes of code
}

/**************************************************************************************/

unsigned int ui_hook( unsigned int msg ) {
   struct sms_received *sms;
   unsigned char *sms_tpdu;
   unsigned int tp_pid, tp_dcs, tp_udhi, tp_udl, tp_udhl, udh_skip;
   register unsigned int tone_id = tone_sms_ascending_id;// load this constant in a register
   
   switch(msg) {
      case MSG_SMS_RECEIVED:
         // Try to generate tone from the SMS
         sms = (struct sms_received*)msg_argv[0];
         if( sms->type != SMS_TYPE_INCOMING ) break;     // quit, delivery report
         sms_tpdu = sms->data + sms_received_padding;
         
         // Parse SMS TPDU fields
         tp_pid = sms_tpdu[1];
         if( tp_pid != 0x00 ) break;                     // quit, unknown TP-PID
         tp_dcs = sms_tpdu[2] & 0xFC;
         if( tp_dcs > 0x10 && tp_dcs != 0xF0 ) break;    // quit, unsupported TP-DCS
         tp_udhi = ( sms_tpdu[0] & 0x40 );
         if( tp_udhi ) {   // Calculate count of letters used for User Data Header
            tp_udhl = sms_tpdu[23];
            udh_skip = ( ( tp_udhl + 1 ) * 8 + 6 ) / 7;
         } else {
            udh_skip = 0;
         }
         tp_udl = sms_tpdu[3];
         if( tp_udl <= udh_skip ) break;                 // quit, empty message
         
         build_morse_tone( &sms_tpdu[23], tp_udl, udh_skip );
         break;
         
      case MSG_PLAY_SYSTEM_TONE:
         // If tone was successfully generated, replace the built-in "Ascending" tone with it
         if( tone != NULL ) {
            if( msg_argv[0] == tone_id ) {
               if( tone->ptr != NULL ) {
                  msg = MSG_NOP;                         // Previous morse code currently playing!
               } else {
                  tone->ptr = tone->data;                // "Now playing"
                  msg_argv[0] = (unsigned int)tone;
                  msg_argv[2] = tone_id;
                  msg_argc = 3;
                  msg = MSG_PLAY_TONE;
               }
            } else {
               free(tone);                               // Something else than Ascending, discard
               tone = NULL;
            }
         }
         break;

      case 0x0546:   // Comes at the end of the tone(?) - TO BE VERIFIED
         if( msg_argv[1] != 0x7F ) break;
         // FALL-THROUGH
      case MSG_STOP_TONE:
         // Discard generated tone after use
         if( msg_argv[0] == tone_id ) {
            // Discarding tone is important, otherwise next message will not have a tone
            // (thinks in create_morse_tone() that previous morse tone is still playing!)
            free(tone);
            tone = NULL;
         }
         break;
   }
   return msg;
}

/**************************************************************************************/

